/**
 * Copyright (c) 2013-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @providesModule RelayFlowGenerator
 * 
 * @format
 */

'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

const t = require('babel-types');

const {
  FlattenTransform,
  IRVisitor,
  SchemaUtils
} = require('relay-compiler/lib/GraphQLCompilerPublic');
const {
  exactObjectTypeAnnotation,
  exportType,
  lineComments,
  readOnlyArrayOfType,
  readOnlyObjectTypeProperty,
  stringLiteralTypeAnnotation
} = require('./ScalaBabelFactories');
const {
  transformScalarType,
  transformInputType
} = require('./ScalaTypeTransformers');
const { GraphQLNonNull } = require('graphql');

const babelGenerator = require('babel-generator').default;

const { isAbstractType } = SchemaUtils;

function generate(node, customScalars, inputFieldWhiteList) {
  const ast = IRVisitor.visit(node, createVisitor(customScalars || {}, inputFieldWhiteList));

  const code = babelGenerator(ast).code;
  console.log(code);
  return code;
}

function makeProp({ key, schemaName, value, conditional, nodeType, nodeSelections }, customScalars, concreteType) {
  if (nodeType) {
    value = transformScalarType(nodeType, customScalars, selectionsToBabel([Array.from(nodeSelections.values())], customScalars));
  }
  if (schemaName === '__typename' && concreteType) {
    value = stringLiteralTypeAnnotation(concreteType);
  }
  const typeProperty = readOnlyObjectTypeProperty(key, value);
  if (conditional) {
    typeProperty.optional = true;
  }
  return typeProperty;
}

const isTypenameSelection = selection => selection.schemaName === '__typename';
const hasTypenameSelection = selections => selections.some(isTypenameSelection);
const onlySelectsTypename = selections => selections.every(isTypenameSelection);

function selectionsToBabel(selections, customScalars) {
  const baseFields = new Map();
  const byConcreteType = {};

  flattenArray(selections).forEach(selection => {
    const { concreteType } = selection;
    if (concreteType) {
      byConcreteType[concreteType] = byConcreteType[concreteType] || [];
      byConcreteType[concreteType].push(selection);
    } else {
      const previousSel = baseFields.get(selection.key);

      baseFields.set(selection.key, previousSel ? mergeSelection(selection, previousSel) : selection);
    }
  });

  const types = [];

  if (Object.keys(byConcreteType).length && onlySelectsTypename(Array.from(baseFields.values())) && (hasTypenameSelection(Array.from(baseFields.values())) || Object.keys(byConcreteType).every(type => hasTypenameSelection(byConcreteType[type])))) {
    for (const concreteType in byConcreteType) {
      types.push(exactObjectTypeAnnotation([...Array.from(baseFields.values()).map(selection => makeProp(selection, customScalars, concreteType)), ...byConcreteType[concreteType].map(selection => makeProp(selection, customScalars, concreteType))]));
    }
    // It might be some other type then the listed concrete types. Ideally, we
    // would set the type to diff(string, set of listed concrete types), but
    // this doesn't exist in Flow at the time.
    const otherProp = readOnlyObjectTypeProperty('__typename', stringLiteralTypeAnnotation('%other'));
    otherProp.leadingComments = lineComments("This will never be '%other', but we need some", 'value in case none of the concrete values match.');
    types.push(exactObjectTypeAnnotation([otherProp]));
  } else {
    let selectionMap = selectionsToMap(Array.from(baseFields.values()));
    for (const concreteType in byConcreteType) {
      selectionMap = mergeSelections(selectionMap, selectionsToMap(byConcreteType[concreteType].map(sel => _extends({}, sel, {
        conditional: true
      }))));
    }
    const selectionMapValues = Array.from(selectionMap.values()).map(sel => isTypenameSelection(sel) && sel.concreteType ? makeProp(_extends({}, sel, { conditional: false }), customScalars, sel.concreteType) : makeProp(sel, customScalars));
    types.push(exactObjectTypeAnnotation(selectionMapValues));
  }

  if (types.length === 0) {
    return exactObjectTypeAnnotation([]);
  }

  return types.length > 1 ? t.unionTypeAnnotation(types) : types[0];
}

function mergeSelection(a, b) {
  if (!a) {
    return _extends({}, b, {
      conditional: true
    });
  }
  return _extends({}, a, {
    nodeSelections: a.nodeSelections ? mergeSelections(a.nodeSelections, b.nodeSelections) : null,
    conditional: a.conditional && b.conditional
  });
}

function mergeSelections(a, b) {
  const merged = new Map();
  for (const [key, value] of a.entries()) {
    merged.set(key, value);
  }
  for (const [key, value] of b.entries()) {
    merged.set(key, mergeSelection(a.get(key), value));
  }
  return merged;
}

function isPlural({ directives }) {
  const relayDirective = directives.find(({ name }) => name === 'relay');
  return relayDirective != null && relayDirective.args.some(({ name, value }) => name === 'plural' && value.value);
}

function createVisitor(customScalars, inputFieldWhiteList) {
  return {
    leave: {
      Root(node) {
        const statements = [];
        if (node.operation !== 'query') {
          statements.push(generateInputVariablesType(node, customScalars, inputFieldWhiteList));
        }
        statements.push(exportType(`${node.name}Response`, selectionsToBabel(node.selections, customScalars)));
        return t.program(statements);
      },

      Fragment(node) {
        let selections = flattenArray(node.selections);
        const numConecreteSelections = selections.filter(s => s.concreteType).length;
        selections = selections.map(selection => {
          if (numConecreteSelections <= 1 && isTypenameSelection(selection) && !isAbstractType(node.type)) {
            return [_extends({}, selection, {
              concreteType: node.type.toString()
            })];
          }
          return [selection];
        });
        const baseType = selectionsToBabel(selections, customScalars);
        const type = isPlural(node) ? readOnlyArrayOfType(baseType) : baseType;

        return t.program([exportType(node.name, type)]);
      },

      InlineFragment(node) {
        const typeCondition = node.typeCondition;
        return flattenArray(node.selections).map(typeSelection => {
          return isAbstractType(typeCondition) ? _extends({}, typeSelection, {
            conditional: true
          }) : _extends({}, typeSelection, {
            concreteType: typeCondition.toString()
          });
        });
      },
      Condition(node) {
        return flattenArray(node.selections).map(selection => {
          return _extends({}, selection, {
            conditional: true
          });
        });
      },
      ScalarField(node) {
        return [{
          key: node.alias || node.name,
          schemaName: node.name,
          value: transformScalarType(node.type, customScalars)
        }];
      },
      LinkedField(node) {
        return [{
          key: node.alias || node.name,
          schemaName: node.name,
          nodeType: node.type,
          nodeSelections: selectionsToMap(flattenArray(node.selections))
        }];
      },
      FragmentSpread(node) {
        return [];
      }
    }
  };
}

function selectionsToMap(selections) {
  const map = new Map();
  selections.forEach(selection => {
    const previousSel = map.get(selection.key);
    map.set(selection.key, previousSel ? mergeSelection(previousSel, selection) : selection);
  });
  return map;
}

function flattenArray(arrayOfArrays) {
  const result = [];
  arrayOfArrays.forEach(array => result.push(...array));
  return result;
}

function generateInputVariablesType(node, customScalars, inputFieldWhiteList) {
  return exportType(`${node.name}Variables`, exactObjectTypeAnnotation(node.argumentDefinitions.map(arg => {
    const property = t.objectTypeProperty(t.identifier(arg.name), transformInputType(arg.type, customScalars, inputFieldWhiteList));
    if (!(arg.type instanceof GraphQLNonNull)) {
      property.optional = true;
    }
    return property;
  })));
}

const FLOW_TRANSFORMS = [ctx => FlattenTransform.transform(ctx, {})];

module.exports = {
  generate,
  flowTransforms: FLOW_TRANSFORMS
};